#!/usr/bin/env python

# Copyright 2020 Unlimited Robotics

import argparse
import os
import sys
import platform
import subprocess

import rayasdk.constants as constants
from rayasdk.connect import URConnect
from rayasdk.vcs import URVCS, check_simulation_versions, check_sdk_version
from rayasdk.logger import log_error, log_verbose, log_info, set_logger_level
from rayasdk.logger import log_success
from rayasdk.scanner import URScanner
from rayasdk.initializer import URInitializer
from rayasdk.runner import URRunner
from rayasdk.simulator import URSimulator
from rayasdk.killer import URKiller


class URSDK:

    def __init__(self):
        # Check for temp folder
        if not self.init_ursdk_folder():
            return False
        # Check .ur folder
        if not self.check_global_folder():
            log_error(f'Could not create .ur folder')
            return False
        
        self.command_objects = []
        self.command_objects.append(URInitializer())
        self.command_objects.append(URScanner())
        self.command_objects.append(URConnect())
        self.command_objects.append(URRunner())
        self.command_objects.append(URSimulator())
        self.command_objects.append(URKiller())
        self.command_objects.append(URVCS())
        self.init_parser()

    def init_parser(self):
        # Init top parser
        self.argparser = argparse.ArgumentParser(
            description='Unlimited Robotics SDK',
        )
        group = self.argparser.add_mutually_exclusive_group()
        group.add_argument("-v",
                           "--verbose",
                           action="store_true",
                           help="increase output verbosity.")
        group.add_argument("-q",
                           "--quiet",
                           action="store_true",
                           help="don't print on stdout.")
        subparsers = self.argparser.add_subparsers(help='SDK Command',
                                                   dest='command')
        subparsers.required = True
        # Init subparsers
        for obj in self.command_objects:
            obj.init_parser(subparsers)
        # Parse arguments
        self.args, self.unknownargs = self.argparser.parse_known_args()
        # Setup Logger
        set_logger_level(self.args.verbose, self.args.quiet)

    def check_global_folder(self):
        try:
            # create folders that are needed in the .ur path
            os.makedirs(constants.SIMS_APPS, exist_ok=True)
            os.makedirs(constants.SIMS_DATA_APPS, exist_ok=True)
            os.makedirs(constants.SIMS_DATA_APPS_DEVEL, exist_ok=True)
            # if ur_root not exist or is empty clone it
            if not os.path.exists(constants.RAYAENV_DOCKER_UR_ROOT) or len(
                    os.listdir(constants.RAYAENV_DOCKER_UR_ROOT)) == 0:
                log_info('Preparing UR user folder...')
                clone_repo = subprocess.Popen(
                        args=constants.RAYAENV_DOCKER_UR_ROOT_REPO,
                        shell=True,
                        stderr=subprocess.PIPE
                    )
                if clone_repo.returncode == 0:
                    log_success('Done')
                else:
                    _, err = clone_repo.communicate()
                    log_error(err.decode("utf-8"))

        except OSError as exc:
            log_error(exc)
            return False
        return True

    def init_ursdk_folder(self):
        if not os.path.exists(constants.URSDK_TEMP_PATH):
            log_verbose(
                f'Folder {constants.URSDK_TEMP_PATH} does not exists, creating it.')
            try:
                os.mkdir(constants.URSDK_TEMP_PATH)
            except OSError as exc:
                log_error(f'Folder {constants.URSDK_TEMP_PATH} could not be created.')
                return False
            return True
        else:
            log_verbose(f'Folder {constants.URSDK_TEMP_PATH} found.')
            return True

    def run(self):
        try:
            # Check if sdk have an update
            check_sdk_version()
            # Check simulation version on simulator command
            if self.args.command == URSimulator().COMMAND:
                updated = check_simulation_versions()
                if not updated:
                    log_error('RayaOS is not up to date. To update run "rayasdk update"')
                    return False
            # Execute Command
            for obj in self.command_objects:
                if self.args.command == type(obj).COMMAND:
                    return obj.run(self.args, self.unknownargs)
        except Exception as e:
           log_error(str(e))
           return False


def main():
    if platform.system() not in ['Linux', 'Windows', 'Darwin']:
        log_info(f'Platform \'{platform.system()}\' not supported.')
        exit(1)
    ursdk = URSDK()
    ret = ursdk.run()
    if not ret:
        log_verbose('Finished with error.')
        sys.exit(1)
    else:
        sys.exit(0)


if __name__ == "__main__":
    main()
