import platform
import docker
import subprocess
import os

import rayasdk.constants as constants
from rayasdk.messages import *
from rayasdk.logger import log_info, log_error


class DockerHandler():

    def __init__(self):
        self.raya_imgs = {}
        self.docker_client: docker.DockerClient = None
        self.RAYAENV_DOCKER_VERSION = constants.RAYAENV_DOCKER_VERSION_DEFAULT
        self.RAYAENV_DOCKER_IMAGE_NAME = None
        self.check_if_gpu()
        self.check_docker_instalation()
        self.docker_image_version()
    
    
    def check_docker_instalation(self):
        # Check docker installation
        try:
            self.docker_client = docker.from_env()
        except docker.errors.DockerException:
            # maybe required for MAC as well:
            if platform.system() == 'Windows':
                raise Exception(MSG_DOCKER_NOT_RUNNING)
            else:
                raise Exception(MSG_DOCKER_NOT_INSTALLER)
        return True


    def check_if_gpu(self):
        return_code = subprocess.call(
            'docker info | grep "Runtimes:" | grep "nvidia"',
            shell=True,
            stdout=subprocess.PIPE)
        if return_code != 0:
            name = f'unlimitedrobotics/{constants.RAYAENV_DOCKER_IMGNAME}.{constants.RAYAENV_DOCKER_IMG_VERSION_NAME}.prod'
        else:
            name = f'unlimitedrobotics/{constants.RAYAENV_DOCKER_IMGNAME}.{constants.RAYAENV_DOCKER_IMG_VERSION_GPU}.prod'
            
        self.RAYAENV_DOCKER_IMAGE_NAME = name


    def docker_image_version(self):
        # Docker image version
        self.check_docker_instalation()
        images = list()
        for image in self.docker_client.images.list(all=True):
            for subtag in image.tags:
                if subtag.startswith(self.RAYAENV_DOCKER_IMAGE_NAME):
                    try:
                        images.append(subtag.split(':')[1])
                    except:
                        pass
        images.sort()

        if images:
            self.RAYAENV_DOCKER_VERSION = images[-1]

    def __remove_old_images(self, latest_version: str):
        log_info(('Removing old Ra-Ya images ...'))
        for image in self.docker_client.images.list(all=True):
            if image.tags and self.RAYAENV_DOCKER_IMAGE_NAME not in ":".join(
                    image.tags):
                continue
            if any(
                    item.startswith(self.RAYAENV_DOCKER_IMAGE_NAME)
                    for item in image.tags):
                # if the latest image exist do not delete it
                if any(
                        str(item) ==
                        f'{self.RAYAENV_DOCKER_IMAGE_NAME}:{latest_version}'
                        for item in image.tags):
                    continue
                remove_old_image = input(('Found a old raya image with the tags:\n'
                                        f'{image.tags}\n'
                                        'Would you like to delete it?[y/N]\n'))
                if remove_old_image in ['y', 'Y', 'yes', 'Yes', 'YES']:
                    self.docker_client.images.remove(image.id, force=True)
                    log_info((f'\'{image.tags[0]}\' Image removed'))


    def __download_image(self, version: str):
        if self.check_image(version=version):
            log_info(('The Ra-Ya OS image already exist: '
                    f'{self.RAYAENV_DOCKER_IMAGE_NAME}:{version}'))
            return True

        log_info(f'Downloading Ra-Ya OS v{version} (only once)...')
        image_name = f'{self.RAYAENV_DOCKER_IMAGE_NAME}:{version}'
        try:
            if os.system(f'docker pull {image_name}') != 0:
                raise KeyboardInterrupt('')
        except KeyboardInterrupt:
            log_error('Failed to download Ra-Ya OS image. Please try again.')
            return False
        return True


    def download_raya_image(self, version: str):
        self.__remove_old_images(version)
        if not self.__download_image(version=version):
            return False
        return True


    def launch_command(self, output_file='', silent=False):
        if not os.path.exists(constants.RAYAENV_DOCKER_UR_ROOT) or len(
                os.listdir(constants.RAYAENV_DOCKER_UR_ROOT)) == 0:
            log_error(f'\'{constants.RAYAENV_DOCKER_UR_ROOT}\' files does not exist')
            return False
        command = (f'cd {constants.RAYAENV_DOCKER_UR_ROOT}/scripts && '
                f'./run.container.sh {self.RAYAENV_DOCKER_VERSION}')
        docker_command = f'{command} '
        if silent:
            if output_file:
                docker_command += f'> {str(output_file)}'
            else:
                docker_command += f'> /dev/null'
        else:
            if output_file:
                docker_command += f'| tee {str(output_file)}'
        return subprocess.Popen(docker_command, shell=True)


    def launch_app_on_sim(self, app_id, args, debug=False):
        container_name = self.check_running_container()
        execution_route = f'/opt/raya_os/rayadevel/apps/{app_id}/'
        if debug:
            cmd = (f'python3 -m debugpy --wait-for-client --listen 5678 '
                '__main__.py ' + ' '.join(args))
        else:
            cmd = f'python3 __main__.py ' + ' '.join(args)
        command = ('cd /opt/raya_os/rayadevel && '
                'source .bashrc && '
                f'cd {execution_route} && '
                f'{cmd}')

        docker_command = (f'docker exec -it -u rayadevel {container_name} '
                        f'bash -c "{command}"')
        ret = subprocess.call(docker_command, shell=True)


    def kill_apps(self):
        container_name = self.check_running_container()
        log_info(f'Killing all running apps...')
        command = 'pkill -9 -u rayadevel'
        docker_command = (f'docker exec -it {container_name} {command}')
        ret = subprocess.call(docker_command, shell=True)


    def check_running_container(self):
        # Check running container
        container_name = None
        container_dev_name = None

        for container in self.docker_client.containers.list(all=True):
            if str(container.name).startswith(f'{constants.RAYAENV_DOCKER_IMGNAME}'):
                container_name = container.name
            elif container.name == constants.RAYAENV_DOCKER_IMGNAME_DEV:
                container_dev_name = container.name

        if container_name is None and container_dev_name is None:
            raise Exception('The container is not running.')
        elif container_name is not None and container_dev_name is not None:
            raise Exception('Dev and Release container are running, stop one.')

        if container_name:
            return container_name
        elif container_dev_name:
            return container_dev_name
        else:
            return False


    def check_image(self, version: str = None):
        if version is None:
            version = self.RAYAENV_DOCKER_VERSION

        for image in self.docker_client.images.list(all=True):
            if image.tags and self.RAYAENV_DOCKER_IMAGE_NAME not in ":".join(
                    image.tags):
                continue
            if f'{self.RAYAENV_DOCKER_IMAGE_NAME}:{version}' in image.tags:
                return True
        return False
    
    def RayaOS_not_found(self):
        return not self.check_image()
