# Copyright 2020 Unlimited Robotics

import json
import subprocess
import os
from pathlib import Path

import rayasdk.constants as constants
from rayasdk.messages import *
from rayasdk.logger import log_error, log_info, log_warning, log_status
from rayasdk.utils import open_connection_file, validate_app_id_and_folder_name
from rayasdk.container_handlers.docker_handler import DockerHandler


class URRunner:

    COMMAND = 'run'

    def __init__(self):
        self.docker_handler = DockerHandler()

    def init_parser(self, subparser):
        self.parser = subparser.add_parser(
            type(self).COMMAND,
            add_help=False,
            help="connect current raya project to a robot or simulator.")
        self.parser.add_argument(
            '--debug',
            help='runs applications in debug mode',
            action="store_true",
        )

    def run(self, args, unknownargs):
        self.args = args
        self.unknownargs = unknownargs
        # Open connection file
        connection_settings = open_connection_file()

        # Open exec_settings file
        try:
            with open(constants.EXECSETTINGS_PATH, 'r', encoding='utf-8') as f:
                exec_settings = json.load(f)
            app_id = exec_settings[constants.JSON_EXECINFO_APPID]
        except OSError:
            log_error((f'Could not open "{constants.EXECSETTINGS_FILE}" file, '
                       'project not initialized or file not valid.'))
            return False
        except KeyError as key:
            log_error(f'{key} key not found on {constants.EXECSETTINGS_FILE}.')
            return False
        # Check if the parent has the same name that the app id
        if not validate_app_id_and_folder_name(app_id):
            log_error('The name of the folder is different from the app-id.')
            return False

        # Execute app
        try:
            if connection_settings[constants.JSON_EXECINFO_SIM]:
                self.run_app_on_simulation(exec_settings, connection_settings)
            else:
                self.run_app_on_remote(exec_settings, connection_settings)
        except Exception as error:
            log_error(str(error))
            return False
        return True

    def run_app_on_simulation(self, exec_settings, connection_settings):
        app_id = exec_settings[constants.JSON_EXECINFO_APPID]
        origin_path = os.getcwd()

        # Sync the app in the folder of the simulation
        log_status(MSG_SYNC_APP_SIM)
        command = \
            f'rsync --archive --delete {origin_path}/ {constants.SIMS_APPS}/{app_id}'
        subprocess.call(command, shell=True)

        if self.args.debug:
            log_status(MSG_WAIT_DEBUG_CLIENT)

        log_status(MSG_LAUNCHING_APP)
        self.docker_handler.launch_app_on_sim(
                app_id=app_id,
                args=self.unknownargs,
                debug=self.args.debug
            )
        log_status(MSG_APP_FINISHED)

    def run_app_on_remote(self, exec_settings, connection_settings):
        app_id = exec_settings[constants.JSON_EXECINFO_APPID]
        robot_id = connection_settings[constants.JSON_EXECINFO_ROBCONN]\
                                                        [constants.JSON_EXECINFO_ROBID]
        host = connection_settings[constants.JSON_EXECINFO_ROBCONN][constants.JSON_EXECINFO_ROBIP]
        hostname = f'{constants.SSH_USER}@{host}'
        log_status(f'{MSG_SYNC_APP_ROBOT} \'{hostname}\'...')
        
        while True:
            try:
                origin_path = os.getcwd()
                command = (('rsync '
                            '--archive '
                            '--delete '
                            f'--rsh=\'ssh -p{constants.SSH_PORT}\' '
                            f'{origin_path}/ '
                            f'{hostname}:{constants.SSH_APP_FOLDER}/{app_id}'))
                subprocess.run(command, shell=True,
                               capture_output=True).check_returncode()

            except subprocess.CalledProcessError as e:
                if 'REMOTE HOST IDENTIFICATION HAS CHANGED' in str(e.stderr):
                    log_info('')
                    log_warning(
                        ('Host ID have changed, run \'rayasdk connect\' '
                         'again'))
                    command = (('ssh-keygen '
                                f'-f "{Path.home()/".ssh"/"known_hosts"}" '
                                f'-R "[{host}]:{constants.SSH_PORT}"'))
                    subprocess.run(
                        command,
                        shell=True,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL).check_returncode()
                    return
                log_info('')
                log_error('Can not send the app to the robot')
                if 'No route to host' in str(e.stderr):
                    log_info(('No toute to host. Check that the robot is powered'
                           ' on'))
                elif 'Connection refused' in str(e.stderr):
                    log_info(
                        ('Connection refused. Check that the Ra-Ya system is '
                         'running on the robot'))
                else:
                    log_info('Check that the robot is powered on')
                log_info(f'Trying to connect to: {robot_id} / {host}')
                return

            except Exception as error:
                log_error(error)
                return
            
            log_status(MSG_LAUNCHING_APP)
            self.launch_app(app_id, hostname, self.unknownargs)
            log_status(MSG_APP_FINISHED)
            return

    def launch_app(self, app_id, host_url, args):
        execution_route = f'%s%s' % (constants.SSH_APP_FOLDER, app_id)

        if self.args.debug:
            cmd = (f'python3 -m debugpy --listen 5678 --wait-for-client '
                   './__main__.py ')
            cmd += ' '.join(args)
            command = (('ssh -2 -t '
                        '-o LogLevel=QUIET '
                        '-L 5678:localhost:5678 '
                        f'-p {constants.SSH_PORT} {host_url} '
                        f'\'cd {execution_route} && {cmd}\''))
            log_info('Waiting for debug client (VSCode)...')
        else:
            cmd = f'python3 __main__.py ' + ' '.join(args)
            command = (('ssh -t '
                        '-o LogLevel=QUIET '
                        f'-p {constants.SSH_PORT} {host_url} '
                        f'\'cd {execution_route} && {cmd}\''))

        subprocess.call(command, shell=True)
        log_info('')
