# Copyright 2020 Unlimited Robotics
import os
import subprocess
import urllib
import tarfile
import platform
import re

from simple_file_checksum import get_checksum
from time import sleep
from pathlib import Path

import rayasdk.constants as constants
from rayasdk.messages import *
from rayasdk.utils import run_simulator_bridge, MyProgressBar
from rayasdk.logger import log_error, log_info, log_status


class URSimulator:

    COMMAND = 'simulator'

    def __init__(self):
        self.host_os = platform.system()
        self.SIM_BINARY = None
        self.GARYSIM_URL =None
        self.SIM_TARPATH = None
        self.GARYSIM_SHA256 = None
        self.SIMS_HOME = constants.SIMS_HOME
        self.GARYSIM_VERSION = self.get_simulation_version()
        self.init_sim_path_os()
    
    
    def init_sim_path_os(self):
        if self.host_os == 'Linux':
            self.SIM_PATH = self.SIMS_HOME / Path(f'raya_simulator_{self.GARYSIM_VERSION}_linux')
            self.SIM_TARPATH = self.SIMS_HOME / Path(f'raya_simulator_{self.GARYSIM_VERSION}_linux.tar.gz')
            self.SIM_BINARY = str(self.SIM_PATH / Path('Linux.x86_64'))

        elif self.host_os == 'Windows':
            self.SIM_PATH = self.SIMS_HOME / Path(f'raya_simulator_{self.GARYSIM_VERSION}_windows')
            self.SIM_TARPATH = self.SIMS_HOME / Path(f'raya_simulator_{self.GARYSIM_VERSION}_windows.tar.gz')
            self.SIM_BINARY = str(self.SIM_PATH / Path('Windows.exe'))

        # TODO: MAC version - M1
        elif self.host_os == 'Darwin':
            self.SIM_PATH = self.SIMS_HOME / Path(f'raya_simulator_{self.GARYSIM_VERSION}_mac')
            self.SIM_TARPATH = self.SIMS_HOME / Path(f'raya_simulator_{self.GARYSIM_VERSION}_mac.tar.gz')
            self.SIM_BINARY = str(self.SIM_PATH / Path('Mac.App'))

    def get_simulation_version(self):
        # Simulator version
        sim_versions = [
                    re.search(r'\d+\.\d+\.\d+', folder_name).group()
                    for folder_name in os.listdir(constants.SIMS_HOME) 
                    if os.path.isdir(os.path.join(constants.SIMS_HOME, folder_name)) and
                        folder_name.startswith('raya_simulator_')
                ]
        sim_versions.sort()
        if sim_versions:
            return sim_versions[-1]
        else:
            return constants.GARYSIM_VERSION_DEFAULT

    def init_parser(self, subparser):
        self.parser = subparser.add_parser(type(self).COMMAND,
                                           help="run the simulator bridge.")
        self.parser.add_argument(
            '--debug',
            help='Runs only the unity simulation for dev on the container',
            action="store_true")


    def run(self, args, unknownargs):
        self.args = args
        self.unknownargs = unknownargs

        if not os.path.exists(self.SIM_BINARY):
            log_error('Simulator not found, run rayasdk update to download it')
            return False

        # If debug flag not set run container
        container_process = None
        if not self.args.debug:
            try:
                log_status(MSG_RAYA_OS_IMAGE)
                container_process = run_simulator_bridge()
                if type(container_process) == bool:
                    return False
                sleep(5)
            except KeyboardInterrupt:
                pass

        # Run unity simulation
        log_status(MSG_UNITY_SIM)
        simulator_process = subprocess.Popen([str(self.SIM_BINARY)] +
                                             self.unknownargs,
                                             stdout=subprocess.PIPE)

        if container_process is not None:
            while True:
                if container_process.poll() is not None:
                    break
            container_process.terminate()
        else:
            simulator_process.wait()

        simulator_process.terminate()
        return True
    
    
    def simulator_not_found(self):
        return not os.path.exists(f'{self.SIM_BINARY}')
    
    def download_simulator(self,response=None):
        if response is not None:
            # Get parameters from VCS API
            for component in response:
                if component['name'] == 'Unity':
                    response = component
            sim_url = response["link"]
            garysim_sha256 = response["checksum"]
            version = response["versionId"]
        else:
            # Set parameters from local constants file
            sim_url = self.GARYSIM_URL
            garysim_sha256 = self.GARYSIM_SHA256
            version = self.GARYSIM_VERSION
        log_info(f'Downloading Ra-Ya simulator v{version}...')
        sim_tarpath = f'{constants.SIMS_HOME}/simulator_{version}.tar.gz'

        try:
            urllib.request.urlretrieve(sim_url, sim_tarpath, MyProgressBar())
        except urllib.error.HTTPError:
            log_error(f'Download error, the url {sim_url} does not exist.')
            return False

        log_info(f'Checking downloaded file...')
        try:
            sha256 = get_checksum(sim_tarpath, algorithm="SHA256")
        except FileNotFoundError:
            log_error('Download error, try again.')
            return False
        if garysim_sha256 != sha256:
            log_error('Download error, try again.')
            return False

        log_info(f'Extracting...')
        simtar = tarfile.open(sim_tarpath)
        simtar.extractall(constants.SIMS_HOME)
        os.remove(sim_tarpath)
        return True
